#include "cf/codec.h"

#include "msgclnt.h"
#include "msg.hpp"

#include <algorithm>

#include <stdlib.h>

// --------------------------------------------------------------

typedef struct SMessageQWorkerArg
{
	SecureSocket * secureSocket;
	MessageQ * messageQ;
} SMessageQWorkerArg;

typedef struct SCallbackWorkerArg
{
	MessageQ * messageQ;
	SCallback callback;
	std::map<std::string, cf::bin> * keyMap;
} SCallbackWorkerArg;

static int messageQworker(void * arg)
{
	SMessageQWorkerArg * inst = (SMessageQWorkerArg *)arg;

	Protocol::Message message;
	cf::size_t size;

	while (true)
	{
		try
		{
			size = 0;

			cf::bin raw = inst->secureSocket->receive();
			size = raw.size();

			message.parse(raw.toString());
		}
		catch (cf::exception & e)
		{
			LOG(e.stackTrace());

			// closed
			if (size == 0)
			{
				// dummy logout
				message.parse(Protocol::Request().logout());
				inst->messageQ->push(message);
				break;
			}
		}

		inst->messageQ->push(message);
	}

	free(inst);

	return 0;
}

static inline SConversation toConversation(const Protocol::Message & message, std::map<std::string, cf::bin> * keyMap)
{
	SConversation c;

	c.sessid = message.get<std::string>(ProtocolType::SESSION_ID);
	c.sensitive = message.get<bool>(ProtocolType::SENSITIVE);
	c.from = message.get<std::string>(ProtocolType::FROM);

	std::string chat = message.get<std::string>(ProtocolType::MESSAGE);
	if (chat.length() > 0)
	{
		crypto aria;
		aria.setKey((*keyMap)[c.sessid]);

		cf::bin decoded = cf::codec::hex::getInstance()->decode(chat);
		cf::bin plain = aria.decrypt(decoded);
		c.message = plain.toString();
	}

	return c;
}

static inline SOpenSession toOpenSession(const Protocol::Message & message, std::map<std::string, cf::bin> * keyMap)
{
	SOpenSession o;

	if (message.mObject.HasKey(ProtocolType::SESSION_KEY))
	{
		o.sessid = message.get<std::string>(ProtocolType::SESSION_ID);
		o.idList = message.getList<std::string>(ProtocolType::ID_LIST);

		std::string encodedKey = message.get<std::string>(ProtocolType::SESSION_KEY);
		cf::bin key = cf::codec::hex::getInstance()->decode(encodedKey);
		(*keyMap)[o.sessid] = key;
	}

	return o;
}

static int callbackWorker(void * arg)
{
	SCallbackWorkerArg * inst = (SCallbackWorkerArg *)arg;

	Protocol::Message message;

	while (true)
	{
		message = inst->messageQ->pop(ProtocolType::LISTEN, false);
		if (message.type() != ProtocolType::NONE)
			inst->callback.onListen(toConversation(message, inst->keyMap));

		message = inst->messageQ->pop(ProtocolType::OPEN_SESSION, false);
		if (message.type() != ProtocolType::NONE)
			inst->callback.onOpenSession(toOpenSession(message, inst->keyMap));

		message = inst->messageQ->pop(ProtocolType::LOGOUT, false);
		if (message.type() != ProtocolType::NONE)
			break;
	}

	free(inst);

	return 0;
}

static inline std::string hashPassword(const std::string & pw)
{
	return cf::codec::hex::getInstance()->encode(crypto().sha256(cf::bin(pw)));
}

static inline std::string getSecret(const std::string & msg, const std::string & sms, const std::string & random)
{
	crypto crypt;
	crypt.setKey(cf::bin(sms + DELIMITER + random));

	return cf::codec::hex::getInstance()->encode(crypt.encrypt(cf::bin(msg)));
}

// --------------------------------------------------------------

void MessageQ::push(const Protocol::Message & message)
{
	mutex.lock();
	messageQ.push_back(message);
	mutex.unlock();
}

Protocol::Message MessageQ::pop(const std::string & requestType, bool isWait)
{
	Protocol::Message msg;

	do
	{
		if (messageQ.size() == 0)
			continue;

		mutex.lock();
		if (messageQ.size() > 0)
		{
			if (messageQ.front().type() == requestType)
			{
				msg = messageQ.front();
				messageQ.pop_front();
				isWait = false;
			}
		}
		mutex.unlock();
	} while (isWait);

	return msg;
}

static bool handshake(cf::network::tcp & socket, std::string & msg, const std::string & type)
{
	Protocol::Message parser;

	socket.send(msg);

	parser.parse(socket.receive().toString());
	if (parser.type() != type)
		THROW_EXCEPTION("invalid message type");

	return parser.get<bool>(ProtocolType::RESULT);
}

// --------------------------------------------------------------

chev::chev()
	: listener(messageQworker),
	  caller(callbackWorker)
{
}

chev::~chev()
{
	secureSocket.close();
}

const std::string & chev::getLastError() const
{
	return error;
}

bool chev::connect(const std::string & host, const unsigned short port)
{
	try
	{
		socket.connect(host, port);
		secureSocket.setSocket(&socket);
		return true;
	}
	catch (cf::exception & e)
	{
		LOG(e.what());
		error = e.stackTrace();
		return false;
	}
}

bool chev::join(const std::string & id, const std::string & pw, const std::string & sms)
{
	try
	{
		std::string random = cf::codec::hex::getInstance()->encode(generateRandom());
		std::string sub = request.join(id, hashPassword(pw), sms);
		std::string secret = getSecret(sub, sms, random);
		std::string msg = request.secretWithRandom(ProtocolType::JOIN, id, secret, random);

		return handshake(socket, msg, ProtocolType::JOIN);
	}
	catch (cf::exception & e)
	{
		LOG(e.what());
		error = e.stackTrace();
		return false;
	}
}

bool chev::login(const std::string & id, const std::string & pw, const std::string & sms)
{
	try
	{
		request.setUserID(id);

		std::string random = cf::codec::hex::getInstance()->encode(generateRandom());
		std::string sub = request.login(hashPassword(pw));
		std::string secret = getSecret(sub, sms, random);
		std::string msg = request.secretWithRandom(ProtocolType::LOGIN, id, secret, random);

		bool succeed = handshake(socket, msg, ProtocolType::LOGIN);
		if (succeed)
		{
			secureSocket.setKey(socket.local().address(), sms);

			SMessageQWorkerArg * arg = (SMessageQWorkerArg *)malloc(sizeof(SMessageQWorkerArg));
			if (!arg)
				return false;

			arg->secureSocket = &secureSocket;
			arg->messageQ = &messageQ;

			listener.start(arg);
		}
		return succeed;
	}
	catch (cf::exception & e)
	{
		LOG(e.what());
		error = e.stackTrace();
		return false;
	}
}

bool chev::sms(const std::string & phone)
{
	try
	{
		return handshake(socket, request.sms(phone), ProtocolType::SMS);
	}
	catch (cf::exception & e)
	{
		LOG(e.what());
		error = e.stackTrace();
		return false;
	}
}

bool chev::addFriend(const std::string & id)
{
	try
	{
		secureSocket.send(request.addFriend(id));

		return messageQ.pop(ProtocolType::ADD_FRIEND).get<bool>(ProtocolType::RESULT);
	}
	catch (cf::exception & e)
	{
		LOG(e.what());
		error = e.stackTrace();
		return false;
	}
}

bool chev::getFriendList(std::vector<SFriend> & friends)
{
	std::vector<SFriend> friendList;

	try
	{
		secureSocket.send(request.getFriendList());

		friends = messageQ.pop(ProtocolType::FRIEND_LIST).getFriendList();

		return messageQ.pop(ProtocolType::GET_FRIEND_LIST).get<bool>(ProtocolType::RESULT);
	}
	catch (cf::exception & e)
	{
		LOG(e.what());
		error = e.stackTrace();
		return false;
	}
}

std::string chev::getSessionID(const std::vector<std::string> & idList)
{
	try
	{
		std::vector<std::string> toList = idList;
		toList.insert(toList.begin(), request.getUserID());

		std::string concat = joinStrings(toList);

		if (sessionMap.find(concat) == sessionMap.end())
		{
			secureSocket.send(request.openSession(toList));

			Protocol::Message message = messageQ.pop(ProtocolType::OPEN_SESSION);

			if (message.mObject.HasKey(ProtocolType::SESSION_KEY))
			{
				std::string sessionID = message.get<std::string>(ProtocolType::SESSION_ID);
				sessionMap[concat] = sessionID;
				std::string encodedKey = message.get<std::string>(ProtocolType::SESSION_KEY);
				cf::bin key = cf::codec::hex::getInstance()->decode(encodedKey);
				keyMap[sessionID] = key;
			}

			bool result = messageQ.pop(ProtocolType::OPEN_SESSION).get<bool>(ProtocolType::RESULT);
			if (!result)
				THROW_EXCEPTION("cannot open session for [" + concat + "]");
		}

		return sessionMap[concat];
	}
	catch (cf::exception & e)
	{
		LOG(e.what());
		error = e.stackTrace();
		return "";
	}
}

bool chev::tell(const SConversation & conversation)
{
	try
	{
		SConversation c = conversation;

		if (c.message.length() > 0)
		{
			crypto aria;
			aria.setKey(keyMap[c.sessid]);

			cf::bin cipher = aria.encrypt(cf::bin(c.message));
			c.message = cf::codec::hex::getInstance()->encode(cipher);
		}

		secureSocket.send(request.tell(c.sessid, c.message, c.sensitive));

		return messageQ.pop(ProtocolType::TELL).get<bool>(ProtocolType::RESULT);
	}
	catch (cf::exception & e)
	{
		LOG(e.what());
		error = e.stackTrace();
		return false;
	}
}

bool chev::listen(const SCallback & callback)
{
	SCallbackWorkerArg * arg = (SCallbackWorkerArg *)malloc(sizeof(SCallbackWorkerArg));
	if (!arg)
		return false;

	arg->messageQ = &messageQ;
	arg->callback = callback;
	arg->keyMap = &keyMap;

	caller.start(arg);

	return true;
}
